
/*]  Copyright (c) 2009-2010, Charles McGarvey  [**************************
**]  All rights reserved.
*
* vi:ts=4 sw=4 tw=75
*
* Distributable under the terms and conditions of the 2-clause BSD license;
* see the file COPYING for a complete text of the license.
*
**************************************************************************/

#include <cstdlib>		// atexit, getenv
#include <functional>
#include <iostream>
#include <string>
#include <unistd.h>		// access

#include <Moof/Log.hh>
#include <Moof/ModalDialog.hh>
#include <Moof/OpenGL.hh>
#include <Moof/Resource.hh>
#include <Moof/Settings.hh>
#include <Moof/Video.hh>

#include "ErrorHandler.hh"
#include "GameLayer.hh"
#include "Main.hh"
#include "TitleLayer.hh"

#if HAVE_CONFIG_H
#include "config.h"
#endif
#include "version.h"


Main::Main(Mf::Settings& settings, Mf::Video& video) :
	Mf::View(settings, video)
{
	Mf::Dispatch& dispatch = Mf::Dispatch::global();
	mNewContextDispatch = dispatch.addTarget("video.newcontext",
											 boost::bind(&Main::setupGL));
	setupGL();
}


void Main::update(Mf::Scalar t, Mf::Scalar dt)
{
	if (children().size() == 0)
	{
		//Mf::logWarning("main view has no children");
		//stop();
		//return;
		addChild(TitleLayer::alloc());
	}

	Mf::View::update(t, dt);
}

void Main::draw(Mf::Scalar alpha) const
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	Mf::View::draw(alpha);
}

bool Main::handleEvent(const Mf::Event& event)
{
	if (Mf::View::handleEvent(event)) return true;

	switch (event.type)
	{
		case SDL_KEYUP:
			if (event.key.keysym.sym == SDLK_f)
			{
				video().toggleFull();
			}
			else if (event.key.keysym.sym == SDLK_l)
			{
				video().toggleCursorGrab();
				video().toggleCursorVisible();
			}
			break;

		case SDL_VIDEORESIZE:
			glViewport(0, 0, event.resize.w, event.resize.h);
			break;

		case SDL_QUIT:
			stop();
			return true;
	}

	return false;
}


std::string Main::getSearchPath()
{
	// Add search paths; they should be searched in this order:
	// 1. YOINK_DATADIR (environment)
	// 2. YOINK_DATADIR (configure)
	
	std::string path;

	char* dataDir = getenv("YOINK_DATADIR");
	if (dataDir)
	{
		path += dataDir;
		path += ":";
	}
	path += YOINK_DATADIR;

	return path;
}

std::string Main::getConfigPath()
{
	// Build the list of config files to search for, in this order:
	// 1. YOINK_DATADIR/yoinkrc
	// 2. /etc/yoinkrc (not for Windows)
	// 3. $HOME/.yoinkrc
	// 4. YOINKRC (environment)

	std::string path("yoinkrc");
	Mf::Resource::getPath(path);

#if !defined(_WIN32)
	path += ":/etc/yoinkrc";
#endif
	path += ":$HOME/.yoinkrc";

	char* configFile = getenv("YOINKRC");
	if (configFile)
	{
		path += ":";
		path += configFile;
	}

	return path;
}


void Main::setupGL()
{
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);
	glShadeModel(GL_SMOOTH);

	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.0);

	glClearColor(0.0, 0.0, 0.0, 1.0);

	//glMatrixMode(GL_PROJECTION);
	//glLoadIdentity();
	//Mf::Scalar ratio = Mf::core.getVideo()->getWidth() /
		//Mf::core.getVideo()->getHeight();
	//gluPerspective(60.0, ratio, 1.0, 250.0);

	//glMatrixMode(GL_MODELVIEW);
}


void Main::printUsage()
{
	std::cout << "Usage: "
			  << PACKAGE" [-h|--help] [-i|--info] [OPTION=VALUE]..."
			  << std::endl
			  << "The alien-smashing action game." << std::endl
			  << std::endl
			  << "Options:" << std::endl
			  << "  -h, --help" << std::endl
			  << "      show this help and exit" << std::endl
			  << "  -i, --info" << std::endl
			  << "      show version and build information" << std::endl
			  << "  detail=1|2|3" << std::endl
			  << "      the level of detail of game scenes" << std::endl
			  << "  fullscreen=true|false" << std::endl
			  << "      if true, uses the entire display" << std::endl
			  << "  framerate=num" << std::endl
			  << "      number of frames to draw per second" << std::endl
			  << std::endl
			  << "See documentation for more options." << std::endl;
}

void Main::printInfo(int argc, char* argv[])
{
	std::string assets;
	std::string datadir;
	std::string config;

	assets.assign(YOINK_DATADIR);
	int accessible = access(assets.c_str(), R_OK);
	if (accessible != 0) assets += " (no access)";

	char* temp = getenv("YOINK_DATADIR");
	if (temp)
	{
		datadir = temp;
		accessible = access(temp, R_OK);
		if (accessible != 0) datadir += " (no access)";
	}

	temp = getenv("YOINKRC");
	if (temp)
	{
		config = temp;
		accessible = access(temp, R_OK);
		if (accessible != 0) config += " (no access)";
	}

	std::cout << "   Executable: " << argv[0] << std::endl
#ifdef YOINK_GITHEAD
			  << "       Commit: "YOINK_GITHEAD << std::endl
#endif
			  << "      Version: "VERSION << std::endl
			  << "        Built: " << COMPILE_TIME << std::endl
			  << "     Compiler: "COMPILER_STRING << std::endl
			  << "       Assets: " << assets << std::endl
			  << "Build options: "
#ifndef HAVE_CLOCK_GETTIME
			  << "-"
#endif
			  << "clock_gettime "
#ifdef NDEBUG
			  << "-"
#endif
			  << "debug "
#ifndef USE_DOUBLE_PRECISION
			  << "-"
#endif
			  << "double-precision "
#ifndef USE_GTK
			  << "-"
#endif
			  << "gtk "
#ifndef PROFILING_ENABLED
			  << "-"
#endif
			  << "profile "
#ifndef USE_QT4
			  << "-"
#endif
			  << "qt4 "
#ifndef USE_THREADS
			  << "-"
#endif
			  << "threads" << std::endl
			  << "      YOINKRC: " << config << std::endl
			  << "YOINK_DATADIR: " << datadir << std::endl;
}


void hello()
{
	std::cout << std::endl << PACKAGE_STRING << std::endl
			  << "Compiled " << __TIME__ " " __DATE__ << std::endl
			  << "Send patches and bug reports to <"
			  PACKAGE_BUGREPORT << ">." << std::endl << std::endl;
}

void goodbye()
{
	std::cout << std::endl << "Goodbye..." << std::endl << std::endl;
}


#include <Moof/Socket.hh>

int main(int argc, char* argv[])
{
	Mf::ResolverTask task("4950", "255.255.255.255", SOCK_DGRAM);
	task.run();

	int i = task.wait();
	Mf::logWarning << "task ended with code: " << i << std::endl;

	std::vector<Mf::SocketAddress>::const_iterator it;
	for (it = task.addresses().begin(); it != task.addresses().end(); ++it)
	{
		Mf::SocketAddress addr = *it;

		Mf::logInfo << "address: " << addr.name() << ":" << addr.port() 
					<< " (" << addr.type() << ")" << std::endl;
	}


	Mf::Packet packet(1000000);

	packet << 5.1234f;
	Mf::logInfo << "packet size: " << packet.size() << std::endl;

	float meh3;
	packet >> meh3;
	Mf::logInfo << "float: " << meh3 << std::endl;
	Mf::logInfo << "packet size: " << packet.size() << std::endl;


	packet << uint16_t(45);
	Mf::logInfo << "packet size: " << packet.size() << std::endl;
	packet << uint64_t(1234567812345679ULL);
	Mf::logInfo << "packet size: " << packet.size() << std::endl;

	packet << true << false << false << true << false << true << true;
	Mf::logInfo << "packet size: " << packet.size() << std::endl;

	std::vector<short> hi;
	hi.push_back(34);
	hi.push_back(-12345);
	hi.push_back(7734);

	for (int a = 0; a < 15900; a++)
	{
		hi.push_back(a);
	}

	packet << hi;
	Mf::logInfo << "packet size: " << packet.size() << std::endl;

	packet << "hello world";
	Mf::logInfo << "packet size: " << packet.size() << std::endl;

	packet << false << false << false << true << false;
	Mf::logInfo << "packet size: " << packet.size() << std::endl;


	Mf::SocketAddress addr("634", "lappy");
	Mf::logInfo << "local addr: " << addr.name() << std::endl;

	for (it = task.addresses().begin(); it != task.addresses().end(); ++it)
	{
		int bcast = 0;

		Mf::SocketAddress addr = *it;
		Mf::Socket sock(addr);

		sock.get(SO_BROADCAST, bcast);
		Mf::logInfo << "bcast: " << bcast << std::endl;

		sock.set(SO_REUSEADDR, 1);
		sock.set(SO_BROADCAST, 1);

		sock.get(SO_BROADCAST, bcast);
		Mf::logInfo << "bcast: " << bcast << std::endl;

		sock.write(packet);
	}


	return 0;


	if (argc > 1)
	{
		std::string arg(argv[1]);
		if (arg == "-h" || arg == "--help")
		{
			Main::printUsage();
			return 0;
		}
		else if (arg == "-i" || arg == "--info")
		{
			Main::printInfo(argc, argv);
			return 0;
		}
	}

	hello();
	atexit(goodbye);

	Mf::Resource::addSearchPaths(Main::getSearchPath());

	Mf::Settings settings(argc, argv, Main::getConfigPath());

	Mf::Log::Level logLevel = Mf::Log::INFO;
	settings.get("loglevel", logLevel);
	Mf::Log::setLevel(logLevel);

	try
	{
		Mf::Video::Attributes attributes(settings);
		attributes.caption = PACKAGE_STRING;
		attributes.icon = Mf::Resource::getPath(PACKAGE".png");

		Mf::Video 	video(attributes);
		Main		mainView(settings, video);

		mainView.run();
		return 0;
	}
	catch (const Mf::Error& error)
	{
		Mf::ModalDialog dialog(Mf::ModalDialog::CRITICAL,
							   PACKAGE_STRING, "Unhandled Exception",
							   getErrorString(error));

		dialog.run();
		return 1;
	}
}

